@@ -1,5 +1,5 @@
 (*
- * Copyright 2003-2009 Savonet team
+ * Copyright 2003-2016 Savonet team
  *
  * This file is part of Ocaml-cry.
  *
@@ -20,21 +20,133 @@
 
  (** OCaml low level implementation of the shout source protocol. *)
 
-type error = 
+type error =
   | Create of exn
   | Connect of exn
   | Close of exn
   | Write of exn
   | Read of exn
   | Busy
+  | Ssl_unavailable
   | Not_connected
   | Invalid_usage
-  | Bad_answer of string option 
+  | Unknown_host of string
+  | Bad_answer of string option
   | Http_answer of int*string*string
 
 exception Error of error
 exception Timeout
 
+type operation = [`Read|`Write|`Both]
+
+type transport = {
+  write: Bytes.t -> int -> int -> int;
+  read: Bytes.t -> int -> int -> int;
+  wait_for : operation -> float -> bool;
+  close: unit -> unit
+}
+
+let ssl = ref None
+
+let register_ssl fn =
+  ssl := Some fn
+
+let () =
+  let callback fn =
+    register_ssl (fun ?bind ~host sockaddr ->
+      let {Cry_ssl.ssl_write;ssl_read;ssl_wait_for;ssl_close} = fn ?bind ~host sockaddr in
+      {write=ssl_write;read=ssl_read;wait_for=ssl_wait_for;close=ssl_close})
+  in
+  Cry_ssl.register callback
+
+let get_ssl () =
+  match !ssl with
+    | Some fn -> fn
+    | None -> raise (Error Ssl_unavailable)
+
+let gethostbyname h =
+  try
+    Unix.gethostbyname h
+  with
+    | Not_found -> raise (Error (Unknown_host h))
+
+let unix_transport ?bind ?timeout sockaddr =
+  let domain =
+    Unix.domain_of_sockaddr sockaddr
+  in
+  let socket =
+    try
+      Unix.socket domain Unix.SOCK_STREAM 0
+    with
+      | e -> raise (Error (Create e))
+  in
+  begin try
+    begin
+     match bind with
+       | None -> ()
+       | Some s ->
+          let bind_addr_inet = (gethostbyname s).Unix.h_addr_list.(0) in
+          (* Seems like you need to bind on port 0 *)
+          let bind_addr = Unix.ADDR_INET(bind_addr_inet, 0) in
+          Unix.bind socket bind_addr ;
+    end
+  with
+    | e ->
+       begin
+         try
+           Unix.close socket
+         with
+           | _ -> ()
+       end;
+       raise (Error (Create e))
+  end;
+  let wait_for operation delay =
+    let events () =
+     match operation with
+       | `Read ->
+            Unix.select [socket] [] [] delay
+       | `Write ->
+            Unix.select [] [socket] [] delay
+       | `Both ->
+            Unix.select [socket] [socket] [] delay
+    in
+    let r,w,_ = events () in
+    match operation with
+      | `Read -> r <> []
+      | `Write -> w <> []
+      | `Both -> r <> [] || w <> []
+  in
+  let transport =
+    { write = Unix.write socket;
+      read = Unix.read socket;
+      wait_for = wait_for;
+      close = (fun () -> Unix.close socket) }
+  in
+  let do_timeout = timeout <> None in
+  let check_timeout () =
+    match timeout with
+      | Some timeout ->
+         (* Block in a select call for [timeout] seconds. *)
+         let _, w, _ = Unix.select [] [socket] [] timeout in
+         if w = [] then
+           raise (Error (Connect Timeout)) ;
+         Unix.clear_nonblock socket;
+         transport
+     | None -> assert false
+  in
+  try
+    if do_timeout then
+      Unix.set_nonblock socket ;
+    Unix.connect socket sockaddr;
+    if do_timeout then
+      Unix.clear_nonblock socket;
+    transport
+  with
+    | Unix.Unix_error(Unix.EINPROGRESS, _, _) -> check_timeout ()
+    | Unix.Unix_error(Unix.EWOULDBLOCK, _, _) when Sys.os_type = "Win32" ->
+        check_timeout ()
+    | e -> raise (Error (Connect e))
+
 let rec string_of_error = 
      function
       | Error (Create e) -> pp "could not initiate a new handler" e
@@ -43,11 +155,13 @@ let rec string_of_error =
       | Error (Read e) -> pp "could not read data from host" e
       | Error (Close e) -> pp "could not close connection" e
       | Error Busy -> "busy"
+      | Error Ssl_unavailable -> "SSL transport is not available"
       | Error Not_connected -> "not connected"
       | Error Invalid_usage -> "invalid usage"
     (*  | Unix.unix_error (code,name,param) ->
           Printf.sprintf "%s in %s(%s)" (Unix.error_message code)
                                          name param *)
+      | Error (Unknown_host h) -> Printf.sprintf "Unknown host: %s" h
       | Timeout -> "connection timeout"
       | Error (Bad_answer s) -> 
            Printf.sprintf "bad answer%s" 
@@ -62,11 +176,19 @@ and
   pp s e =
     Printf.sprintf "%s: %s" s (string_of_error e)
 
-type protocol = Icy | Http
+type verb = Put | Post | Source
+
+let string_of_verb = function
+  | Put -> "PUT"
+  | Post -> "POST"
+  | Source -> "SOURCE"
+
+type protocol = Icy | Http of verb | Https of verb
 
 let string_of_protocol = function
-  | Icy  -> "icy"
-  | Http -> "http"
+  | Icy     -> "icy"
+  | Http v  -> Printf.sprintf "http(%s)" (string_of_verb v)
+  | Https v -> Printf.sprintf "https(%s)" (string_of_verb v)
 
 type content_type = string
 
@@ -77,33 +199,44 @@ let mpeg = "audio/mpeg"
 let content_type_of_string s = s
 let string_of_content_type x = x
 
+type mount =
+  | Icy_id of int
+  | Icecast_mount of string
+
 type connection =
   {
-    mount        : string;
+    mount        : mount;
     user         : string;
     password     : string;
     host         : string;
     port         : int;
+    chunked      : bool;
     content_type : content_type;
     protocol     : protocol;
     headers      : (string,string) Hashtbl.t
   }
 
+let string_of_mount = function
+  | Icy_id id -> Printf.sprintf "\"sid\": %d" id
+  | Icecast_mount mount -> Printf.sprintf "\"mount\": %S" mount
+
 let string_of_connection c = 
   Printf.sprintf 
-    "{ \"mount\": %S,\n\
+    "{ %s,\n\
        \"user\":  %S,\n\
        \"password\": %S,\n\
        \"host\": %S,\n\
        \"port\": %d,\n\
+       \"chunked\": %b,\n\
        \"content_type\": %S,\n\
        \"protocol\": %S,\n\
        \"headers\": { %s } }"
-     c.mount
+     (string_of_mount c.mount)
      c.user
      c.password
      c.host
      c.port
+     c.chunked
      (string_of_content_type c.content_type)
      (string_of_protocol c.protocol)
      (Hashtbl.fold (fun x y z -> Printf.sprintf "%S: %S,\n%s" x y z)
@@ -115,8 +248,8 @@ type metadata = (string,string) Hashtbl.t
 
 (* Metadata socket is only present if icy_cap is true *)
 type connection_data = 
-  { connection      : connection;
-    data_socket     : Unix.file_descr }
+  { connection : connection;
+    transport : transport }
 
 type status_priv = PrivConnected of connection_data | PrivDisconnected 
 
@@ -124,12 +257,12 @@ type status = Connected of connection_data | Disconnected
 
 type t = 
   { 
-    ipv6               : bool;
     timeout            : float;
     connection_timeout : float option;
     bind               : string option;
     mutable icy_cap    : bool;
-    mutable status     : status_priv
+    mutable status     : status_priv;
+    mutable chunk_cap  : bool;
   }
 
 let get_connection_data x =
@@ -137,49 +270,42 @@ let get_connection_data x =
     | PrivConnected d -> d
     | PrivDisconnected -> raise (Error Not_connected)
 
-let create_socket ?(ipv6=false) ?bind () = 
-  let domain = if ipv6 then Unix.PF_INET6 else Unix.PF_INET in
-  let socket = 
-    try
-      Unix.socket domain Unix.SOCK_STREAM 0
-    with
-      | e -> raise (Error (Create e)) 
-  in
-  try
-    begin
-     match bind with
-       | None -> ()
-       | Some s ->
-          let bind_addr_inet = (Unix.gethostbyname s).Unix.h_addr_list.(0) in
-          (* Seems like you need to bind on port 0 *)
-          let bind_addr = Unix.ADDR_INET(bind_addr_inet, 0) in
-          Unix.bind socket bind_addr ;
-    end;
-    socket
-  with
-    | e -> 
-       begin
-         try
-           Unix.close socket
-         with
-           | _ -> ()
-       end;
-       raise (Error (Create e))
-
-let create ?(ipv6=false) ?bind ?connection_timeout ?(timeout=30.) () = 
+let create ?bind ?connection_timeout ?(timeout=30.) () = 
   { 
-    ipv6               = ipv6;
     timeout            = timeout;
     connection_timeout = connection_timeout;
     bind               = bind;
     icy_cap            = false;
-    status             = PrivDisconnected
+    status             = PrivDisconnected;
+    chunk_cap          = false;
   }
 
+let write_data ~timeout transport request =
+  let request = Bytes.of_string request in
+  let len = Bytes.length request in
+  let rec write ofs =
+    if not (transport.wait_for `Write timeout) then
+      raise Timeout;
+    let rem = len - ofs in
+    if rem > 0 then
+      let ret = transport.write request ofs rem in
+      if ret = 0 then
+        raise (Failure "connection closed.") ;
+      if ret < rem then
+        write (ofs+ret)
+  in
+  try
+    write 0
+  with
+    | e -> raise (Error (Write e))
+
 let close x =
     try
       let c = get_connection_data x in
-      Unix.close c.data_socket ;
+      if x.chunk_cap then
+        write_data ~timeout:x.timeout c.transport "0\r\n\r\n";
+      c.transport.close ();
+      x.chunk_cap <- false;
       x.icy_cap <- false;
       x.status <- PrivDisconnected
     with
@@ -198,7 +324,7 @@ let audio_info
   let infos = Hashtbl.create 10 in
   let f m x y = 
     match y with
-      | Some v -> Hashtbl.add infos x (m v)
+      | Some v -> Hashtbl.replace infos x (m v)
       | None -> ()
   in
   f string_of_int "channels" channels;
@@ -207,33 +333,43 @@ let audio_info
   f string_of_float "quality" quality;
   infos
 
+let normalize_mount mount =
+  match mount with
+    | Icecast_mount mount -> Icecast_mount
+        begin
+          match mount with
+            | "" -> "/"
+            | mount ->
+                if mount.[0] = '/' then
+                  mount
+                else
+                  Printf.sprintf "/%s" mount
+        end
+    | _ -> mount
+
 let connection 
     ?(user_agent) ?(name) ?(genre) 
     ?(url) ?(public) ?(audio_info) 
     ?(description) ?(host="localhost") 
-    ?(port=8000) ?(password="hackme")
-    ?(protocol=Http) ?(user="source")
-    ~mount ~content_type () =
+    ?(port=8000) ?(chunked=false)
+    ?(password="hackme") ?(protocol=(Http Source))
+    ?(user="source") ~mount ~content_type () =
   let headers = Hashtbl.create 10 in
   let public = 
     match public with
       | Some x -> if x then Some "1" else Some "0"
       | None -> None
   in
-  let mount = 
-    if mount.[0] = '/' then 
-      mount 
-    else
-      Printf.sprintf "/%s" mount
-  in 
+  let mount = normalize_mount mount in 
   let f (x,y) = 
     match y with
-      | Some v -> Hashtbl.add headers x v 
+      | Some v -> Hashtbl.replace headers x v 
       | None -> ()
   in
   let x = 
     match protocol with
-      | Http -> 
+      | Http _
+      | Https _ -> 
          let audio_info =
            match audio_info with
              | Some x ->
@@ -271,6 +407,7 @@ let connection
     host = host;
     port = port;
     user = user;
+    chunked = chunked;
     password = password;
     mount = mount;
     content_type  = content_type;
@@ -287,9 +424,9 @@ let encode64 s =
   let extra = String.length s mod 3 in
   let s = match extra with 1 -> s ^ "\000\000" | 2 -> s ^ "\000" | _ -> s in
   let n = String.length s in
-  let dst = String.create (4 * (n/3)) in
+  let dst = Bytes.create (4 * (n/3)) in
     for i = 0 to n/3 - 1 do
-      let (:=) j v = dst.[i*4+j] <- digit.[v] in
+      let (:=) j v = Bytes.set dst (i*4+j) digit.[v] in
       let c j = int_of_char s.[i*3+j] in
       let c0 = c 0 and c1 = c 1 and c2 = c 2 in
         0 := c0 lsr 2 ;
@@ -298,11 +435,11 @@ let encode64 s =
         3 := c2 land 63
     done ;
     if extra = 1 then begin
-      dst.[4*(n/3)-2] <- '=' ;
-      dst.[4*(n/3)-1] <- '='
+      Bytes.set dst (4*(n/3)-2) '=' ;
+      Bytes.set dst (4*(n/3)-1) '='
     end else if extra = 2 then
-      dst.[4*(n/3)-1] <- '=' ;
-    dst
+      Bytes.set dst (4*(n/3)-1) '=' ;
+    Bytes.to_string dst
 
 (* URL encoding/decoging according to RFC 1738, RFC 1630.
  * Borrowed from ocamlnet. *)
@@ -314,10 +451,10 @@ let to_hex2 =
        '8'; '9'; 'A'; 'B'; 'C'; 'D'; 'E'; 'F' |]
   in
     fun k ->
-      let s = String.create 2 in
-        s.[0] <- hex_digits.( (k lsr 4) land 15 ) ;
-        s.[1] <- hex_digits.( k land 15 ) ;
-        s
+      let s = Bytes.create 2 in
+        Bytes.set s 0 hex_digits.( (k lsr 4) land 15 ) ;
+        Bytes.set s 1 hex_digits.( k land 15 ) ;
+        Bytes.to_string s
 
 let url_encode s = 
  let rec do_url_encode s s' =
@@ -345,48 +482,12 @@ let url_encode s =
  do_url_encode s ""
 
 let http_header =
-  Printf.sprintf "SOURCE %s HTTP/1.0\r\n%s\r\n\r\n"
+  Printf.sprintf "%s %s HTTP/1.%d\r\n%s\r\n\r\n"
 
 let get_auth user password = 
   Printf.sprintf "Basic %s" (encode64 (user ^ ":" ^ password))
 
-(* Returns [false] if delay has been expired
- * while waiting for the requested operation. *)
-let wait_for operation delay s = 
-  let events () =
-   match operation with 
-     | `Read -> 
-          Unix.select [s] [] [] delay
-     | `Write ->
-          Unix.select [] [s] [] delay
-     | `Both ->
-          Unix.select [s] [s] [] delay
-  in
-  let r,w,_ = events () in
-  match operation with
-    | `Read -> r <> []
-    | `Write -> w <> []
-    | `Both -> r <> [] || w <> []
-
-let write_data ~timeout socket request =
-  let len = String.length request in
-  let rec write ofs =
-    if not (wait_for `Write timeout socket) then
-      raise Timeout ; 
-    let rem = len - ofs in
-    if rem > 0 then
-      let ret = Unix.write socket request ofs rem in
-      if ret = 0 then
-        raise (Failure "connection closed.") ;
-      if ret < rem then
-        write (ofs+ret)
-  in
-  try
-    write 0
-  with
-    | e -> raise (Error (Write e))
-
-let buf = String.create 1024
+let buf = Bytes.create 1024
 
 (** Read and split data. 
   * There should always be at least
@@ -394,34 +495,36 @@ let buf = String.create 1024
   * If not, something bad happened. *)
 (* TODO: review all reading code as this 
  * seems a bit ad-hoc/naive.. *)
-let read_data ~timeout socket =
-  if not (wait_for `Read timeout socket) then
-    raise Timeout ;
+let read_data ~timeout transport =
+  if not (transport.wait_for `Read timeout) then
+    raise Timeout;
   try
-     let ret = Unix.read socket buf 0 1024 in
+     let ret = transport.read buf 0 1024 in
      (* split data *)
-     let buf = String.sub buf 0 ret in
+     let buf = Bytes.sub buf 0 ret in
      let rec f pos l = 
        try
-         let x = String.index_from buf pos '\n' in
-         f (x+1) (String.sub buf pos (x-pos-1)::l)
+         let x = Bytes.index_from buf pos '\n' in
+         f (x+1) (Bytes.sub buf pos (x-pos-1)::l)
        with
          | Invalid_argument _ 
          | Not_found -> 
              if pos < ret then 
-               String.sub buf pos (ret-pos) :: l 
+               Bytes.sub buf pos (ret-pos) :: l 
              else 
                l
      in
      let ret = f 0 [] in
      if List.length ret = 0 then
-       [""]
+       [Bytes.of_string ""]
      else
        List.rev (f 0 [])
   with
      | e -> raise (Error (Read e)) 
 
-let header_string source =
+let header_string headers source =
+    let unique_headers = Hashtbl.create 10 in
+    Hashtbl.iter (Hashtbl.replace unique_headers) headers;
     (* Icy headers are of the form: %s:%s *)
     let sep = 
       if source.protocol = Icy then
@@ -442,7 +545,7 @@ let header_string source =
     let f x y z =
       Printf.sprintf "%s:%s%s" x sep y :: z
     in
-    let headers = Hashtbl.fold f source.headers 
+    let headers = Hashtbl.fold f unique_headers 
                                (f content_label 
                                   source.content_type []) 
     in
@@ -458,45 +561,85 @@ let parse_http_answer s =
     | Scanf.Scan_failure s -> raise (Error (Bad_answer (Some s)))
     | _ -> raise (Error (Bad_answer None))
 
-let connect_http c socket source = 
+let resolve_host host =
+  (Unix.gethostbyname host).Unix.h_addr_list.(0)
+
+let add_host_header ?(force=false) headers host port =
+  try
+    ignore(resolve_host host);
+    if force then Hashtbl.replace headers "Host" ""
+  with
+    | Not_found ->
+        let host = if port = 80 then host else
+          Printf.sprintf "%s:%d" host port
+        in
+        Hashtbl.replace headers "Host" host
+
+let http_path_of_mount = function
+  | Icecast_mount mount -> mount
+  | _ -> raise (Error Invalid_usage)
+
+let connect_http c transport source verb = 
   let auth = get_auth source.user source.password in 
   try
-    Hashtbl.add source.headers "Authorization" auth;
-    let headers = header_string source in
-    let request = http_header source.mount headers in
-    write_data ~timeout:c.timeout socket request;
+    let http_version = if source.chunked then 1 else 0 in
+    let headers = Hashtbl.copy source.headers in
+    Hashtbl.replace headers "Authorization" auth;
+    add_host_header ~force:(http_version = 1) headers source.host source.port;
+    if source.chunked then Hashtbl.replace headers "Transfer-Encoding" "chunked";
+    let headers = header_string headers source in
+    let request = http_header (string_of_verb verb) (http_path_of_mount source.mount)
+      http_version headers
+    in
+    write_data ~timeout:c.timeout transport request;
     (** Read input from socket. *)
-    let ret = read_data ~timeout:c.timeout socket in 
-    let (v,code,s) = parse_http_answer (List.hd ret) in
+    let ret = read_data ~timeout:c.timeout transport in 
+    let (v,code,s) = parse_http_answer (Bytes.to_string (List.hd ret)) in
     if code < 200 || code >= 300 then
       raise (Error (Http_answer (code,s,v)));
+    c.chunk_cap <- v = "1.1";
     c.icy_cap <- true;
     c.status <- 
       PrivConnected { connection = source;
-                      data_socket = socket }
+                      transport = transport }
   with
     | e -> 
-       begin
-        try
+       begin try
           close c
-        with
-          | _ -> ()
+        with _ -> ()
        end ;
        raise e
 
-let connect_icy c socket source = 
-  let request = Printf.sprintf "%s\r\n" source.password in
+let icy_id_of_mount = function
+  | Icy_id id -> id
+  | _ -> raise (Error Invalid_usage)
+
+let connect_icy c transport source = 
+  let password =
+    let user =
+      match source.user with
+        | "" -> ""
+        | user -> Printf.sprintf "%s:" user
+     in
+     let id =
+       match icy_id_of_mount source.mount with
+         | 1 -> ""
+         | id -> Printf.sprintf ":#%d" id
+     in
+     Printf.sprintf "%s%s%s" user source.password id
+  in
+  let request = Printf.sprintf "%s\r\n" password in
   try
-    write_data ~timeout:c.timeout socket request;
+    write_data ~timeout:c.timeout transport request;
     (** Read input from socket. *)
-    let ret = read_data ~timeout:c.timeout socket in
+    let ret = read_data ~timeout:c.timeout transport in
     let f s =
       if s.[0] <> 'O' && s.[1] <> 'K' then
         raise (Error (Bad_answer (Some s)));
     in
     begin
      try
-       Scanf.sscanf (List.hd ret) "%[^\r^\n]" f
+       Scanf.sscanf (Bytes.to_string (List.hd ret)) "%[^\r^\n]" f
      with
        | Scanf.Scan_failure s -> raise (Error (Bad_answer (Some s)))
        | Error _ as e -> raise e
@@ -506,24 +649,24 @@ let connect_icy c socket source =
     let ret = 
       match ret with
         | x :: y :: _ -> y
-        | _ -> List.hd (read_data ~timeout:c.timeout socket)
+        | _ -> List.hd (read_data ~timeout:c.timeout transport)
     in
     let f s =
       c.icy_cap <- true
     in
     begin
      try
-       Scanf.sscanf ret "icy-caps:%[1]" f
+       Scanf.sscanf (Bytes.to_string ret) "icy-caps:%[1]" f
      with
        | Scanf.Scan_failure s -> ()
     end;
     (* Now Write headers *)
-    let headers = header_string source in
+    let headers = header_string source.headers source in
     let request = Printf.sprintf "%s\r\n\r\n" headers in
-    write_data ~timeout:c.timeout socket request;
+    write_data ~timeout:c.timeout transport request;
     c.status <- 
       PrivConnected { connection = source;
-                      data_socket = socket }
+                      transport = transport }
   with
     | e ->
        begin
@@ -534,62 +677,45 @@ let connect_icy c socket source =
        end ;
        raise e
 
-(** This function does *not* close the socket in case of error.. *)   
-let connect_socket ?timeout socket host port = 
-  let do_timeout = timeout <> None in
-  let check_timeout () =
-    match timeout with
-      | Some timeout ->
-         (* Block in a select call for [timeout] seconds. *)
-         let _, w, _ = Unix.select [] [socket] [] timeout in
-         if w = [] then
-           raise (Error (Connect Timeout)) ;
-         Unix.clear_nonblock socket
-       | None -> assert false
-  in
-  try
-    if do_timeout then
-      Unix.set_nonblock socket ;
-    Unix.connect 
-      socket 
-      (Unix.ADDR_INET((Unix.gethostbyname host).Unix.h_addr_list.(0),port)) ;
-    if do_timeout then
-      Unix.clear_nonblock socket
-  with
-    | Unix.Unix_error(Unix.EINPROGRESS, _, _) -> check_timeout ()
-    | Unix.Unix_error(Unix.EWOULDBLOCK, _, _) when Sys.os_type = "Win32" ->
-        check_timeout ()
-    | e -> raise (Error (Connect e))
- 
 let connect c source =
   if c.status <> PrivDisconnected then
     raise (Error Busy);
-  let port = 
-    if source.protocol = Icy then source.port+1 else source.port 
+  let port =
+    if source.protocol = Icy then source.port+1 else source.port
   in
-  let socket = create_socket ~ipv6:c.ipv6
-                             ?bind:c.bind ()
+  let sockaddr =
+    Unix.ADDR_INET(resolve_host source.host,port)
+  in
+  let transport = match source.protocol with
+    | Icy
+    | Http _ -> unix_transport ?bind:c.bind sockaddr
+    | Https _ -> (get_ssl ()) ~host:source.host sockaddr
   in
   try
-    connect_socket ?timeout:c.connection_timeout socket source.host port;
     (* We do not know icy capabilities so far.. *)
     c.icy_cap <- false;
     match source.protocol with
-      | Http -> connect_http c socket source
-      | Icy -> connect_icy c socket source
+      | Http verb
+      | Https verb -> connect_http c transport source verb
+      | Icy -> connect_icy c transport source
   with
     | e -> 
-       begin
-        try
-         Unix.close socket
-        with
+       begin try
+         transport.close ()
+       with
           | _ -> ()
-       end; 
-       raise e
+       end; raise e
 
-let http_meta_request = 
+let http_meta_request mount charset meta headers = 
+  let unique_headers = Hashtbl.create 10 in
+  Hashtbl.iter (Hashtbl.replace unique_headers) headers;
+  let header = 
+     Hashtbl.fold (Printf.sprintf "%s: %s\r\n%s")
+                   unique_headers ""
+  in
   Printf.sprintf 
     "GET /admin/metadata?mode=updinfo&mount=%s%s%s HTTP/1.0\r\n%s\r\n"
+     mount charset meta header
 
 let icy_meta_request = 
   Printf.sprintf 
@@ -597,49 +723,34 @@ let icy_meta_request =
 
 let manual_update_metadata 
        ~host ~port ~protocol ~user ~password 
-       ~mount ?(connection_timeout=5.) ?(timeout=30.) ?headers ?(ipv6=false)
+       ~mount ?(connection_timeout=5.) ?(timeout=30.) ?headers
        ?bind ?charset m =
-  let mount = 
-    if mount.[0] <> '/' then
-      "/" ^ mount
-    else
-      mount
-  in 
+  let mount = normalize_mount mount in 
   let headers = 
     match headers with
-      | Some x -> x
+      | Some x -> Hashtbl.copy x
       | None   -> Hashtbl.create 0
   in
-  let socket = create_socket ~ipv6 ?bind () in
+  let sockaddr =
+    Unix.ADDR_INET(resolve_host host,port)
+  in
+  let transport = match protocol with
+    | Icy
+    | Http _ -> unix_transport ?bind sockaddr
+    | Https _ -> (get_ssl ()) ~host sockaddr
+  in
   let close () = 
     try
-      Unix.close socket
+      transport.close ()
     with
       | e -> raise (Error (Close e))
   in
   try
-    connect_socket ~timeout:connection_timeout socket host port;
-    let user_agent =
-      try
-        Hashtbl.find headers "User-Agent"
-      with
-        | Not_found -> "ocaml-cry"
-    in
-    (** This seems to be needed for shoutcast *)
-    let agent_complement = 
-      if protocol = Icy then
-        " (Mozilla compatible)"
-      else
-        ""
-    in
     let charset = 
       match charset with
         | Some c -> Printf.sprintf "&charset=%s" c
         | None   -> ""
     in
-    let user_agent = 
-      Printf.sprintf "User-Agent: %s%s\r\n" user_agent agent_complement
-    in
     let f x y z =
       if y <> "" then
         Printf.sprintf "%s&%s=%s" z (url_encode x) (url_encode y)
@@ -651,17 +762,42 @@ let manual_update_metadata
     in
     let request = 
       match protocol with
-        | Http ->
-            let headers = 
-              Printf.sprintf "Authorization: %s\r\n%s" (get_auth user password) user_agent
+        | Http _
+        | Https _ ->
+            let mount =
+              match mount with
+                | Icecast_mount mount -> mount
+                | _ -> raise (Error Invalid_usage)
             in
+            Hashtbl.replace headers "Authorization" (get_auth user password);
+            add_host_header headers host port;
             http_meta_request mount charset meta headers
-        | Icy -> icy_meta_request password meta user_agent
+        | Icy ->
+            let sid =
+              match mount with
+                | Icy_id id -> id
+                | _ -> raise (Error Invalid_usage)
+            in
+            let meta =
+              match sid with
+                | 1 -> meta
+                | sid -> Printf.sprintf "%s&sid=%d" meta sid
+            in
+            let user_agent =
+              try
+                Hashtbl.find headers "User-Agent"
+               with
+                | Not_found -> "ocaml-cry"
+            in
+            let user_agent =
+              Printf.sprintf "User-Agent: %s (Mozilla compatible)\r\n" user_agent
+            in
+            icy_meta_request password meta user_agent
     in
-    write_data ~timeout socket request;
+    write_data ~timeout transport request;
     (** Read input from socket. *)
-    let ret = read_data ~timeout socket in
-    let (v,code,s) = parse_http_answer (List.hd ret) in
+    let ret = read_data ~timeout transport in
+    let (v,code,s) = parse_http_answer (Bytes.to_string (List.hd ret)) in
     if code <> 200 then
       raise (Error (Http_answer (code,s,v)));
     close ()
@@ -697,5 +833,13 @@ let update_metadata ?charset c m =
 
 let send c x =
   let d = get_connection_data c in
-  write_data ~timeout:c.timeout d.data_socket x
-
+  if c.chunk_cap then
+   begin
+    if x <> "" then
+      let x =
+        Printf.sprintf "%X\r\n%s\r\n" (String.length x) x
+       in
+       write_data ~timeout:c.timeout d.transport x
+   end
+  else 
+    write_data ~timeout:c.timeout d.transport x
